#ifndef __MD_LIB_FIELDMAP_H__
#define __MD_LIB_FIELDMAP_H__

#include <map>
#include <set>
#include <stdexcept>
#include <string>
#include <vector>

#include <MD/private/prefix.h>
#include <MD/MD_sys_types.h>
#include <MD/MD_Types.h>
#include <MD/MD_FieldDefs.h>
#include <boost/thread/mutex.hpp>

// Enabling this define mimics the behaviour of the FieldMap prior to the changes (2017) to allow alias of field names
// NOTE: It does not do any of the extra validation checks either to maintain the original behaviour
#define MD_NO_FIELD_NAME_ALIAS 1

  namespace MD {
    namespace FIELDS {
      
      typedef uint16_t CompactID;

      struct field_rec {
        MD::FIELDS::FieldID id;
        MD::FIELDS::CompactID compact;
        MD::DATATYPES::TypeID type;
        std::string name;
       
        bool operator< (const field_rec & rhs) const {
          return this->id < rhs.id;
        }
      };
      
    }


  class MD_FieldMap
  {
    public:
    typedef std::set <MD::FIELDS::field_rec> data_type;
    typedef std::vector <data_type::const_iterator> id_index_type;
    typedef std::vector <data_type::const_iterator> compact_index_type;
    typedef std::map <std::string, data_type::const_iterator> name_index_type;
    typedef data_type::const_iterator const_iterator;

    MD_LIB_API static int load_file(const std::string &filename);
    MD_LIB_API static const MD_FieldMap & instance();
    MD_LIB_API static void destroy();

    // This will repopulate the data structures from field_map.txt or
    // FieldDefinition.xml. DO NOT call this unless you can be sure no other
    // threads/functions are using iterators to the FieldMap as this call will
    // INVALIDATE any iterators created before the reload.
    MD_LIB_API static void reload(const std::string & field_def = "");
    MD_LIB_API static void clear_data_structures(MD_FieldMap * instance);
    MD_LIB_API static void reload_xml(const std::string & file);
    MD_LIB_API static void reload_text(const std::string & file);

    // Search for field_rec structures in the field map by id or
    // name.  If the field_rec is found, they return a
    // const_iterator to that field_rec.  If the field_rec is not
    // found, they return an iterator to end().
    MD_LIB_API MD_FieldMap::const_iterator findById (MD::FIELDS::FieldID id) const;
    MD_LIB_API MD_FieldMap::const_iterator findByCompact (MD::FIELDS::CompactID cid) const;
    MD_LIB_API MD_FieldMap::const_iterator findByName (const std::string & name) const;

    // Convenience methods to translate an id to a type or a name,
    // and a name to a type or id.  These methods throw a
    // std::domain_error exception if an id or a name is not found
    // in the Field Map.
    MD_LIB_API MD::DATATYPES::TypeID   idToType   (MD::FIELDS::FieldID id) const;
    MD_LIB_API const std::string &     idToName   (MD::FIELDS::FieldID id) const;
    MD_LIB_API MD::FIELDS::FieldID     nameToId   (const std::string & name) const;
    MD_LIB_API MD::DATATYPES::TypeID nameToType   (const std::string & name) const;

    // const_iterator access to MD_FieldMap
    MD_FieldMap::const_iterator begin() const { return m_fields.begin(); }
    MD_FieldMap::const_iterator end()   const { return m_fields.end();   }

    size_t size() const { return m_fields.size(); }

    private:
    MD_FieldMap();
    ~MD_FieldMap();
    MD_FieldMap(const MD_FieldMap & not_copyable);
    MD_FieldMap & operator=(const MD_FieldMap & not_copyable);

    static MD_FieldMap * checked_get_fieldmap();
        
    void load();
    void insert_field(const MD::FIELDS::field_rec & rec);
    void load_field_map(const std::string & location);
    void load_field_def(const std::string & location);
    void load_helper();

    data_type              m_fields;
    id_index_type          m_id_index;
    compact_index_type     m_compact_index;
    name_index_type        m_name_index;
    MD::FIELDS::FieldID    m_max_id;
    MD::FIELDS::CompactID  m_max_compact;
    static MD_FieldMap    *m_instance;
    static boost::mutex m_lock;
  };


  // Implementation of inline methods goes here, to keep the class
  // declaration clean.

  inline MD_FieldMap::const_iterator
    MD_FieldMap::findById (MD::FIELDS::FieldID id) const {
    if (id <= m_max_id) {
      return m_id_index[id];
    } else {
      return end();
    }
  }

  inline MD_FieldMap::const_iterator
    MD_FieldMap::findByCompact (MD::FIELDS::CompactID cid) const {
    if (cid <= m_max_compact) {
      return m_compact_index[cid];
    } else {
      return end();
    }
  }

  inline MD_FieldMap::const_iterator
    MD_FieldMap::findByName (const std::string & name) const {
    name_index_type::const_iterator i = m_name_index.find(name);
    if (i == m_name_index.end()) { return end(); }
    return i->second;
  }


  inline MD::DATATYPES::TypeID
    MD_FieldMap::idToType (MD::FIELDS::FieldID id) const {
    MD_FieldMap::const_iterator i = findById(id);
    if (i == end()) {
      throw std::domain_error("Invalid id passed to MD_FieldMap::idToType()");
    }
    return i->type;
  }

  inline const std::string &
    MD_FieldMap::idToName (MD::FIELDS::FieldID id) const {
    MD_FieldMap::const_iterator i = findById(id);
    if (i == end()) {
      throw std::domain_error("Invalid id passed to MD_FieldMap::idToName()");
    }
    return i->name;
  }

  inline MD::FIELDS::FieldID
    MD_FieldMap::nameToId (const std::string & name) const {
    MD_FieldMap::const_iterator i = findByName(name);
    if (i == end()) {
      throw std::domain_error("Invalid id passed to MD_FieldMap::nameToId()");
    }
    return i->id;
  }

  inline MD::DATATYPES::TypeID
    MD_FieldMap::nameToType (const std::string & name) const {
    MD_FieldMap::const_iterator i = findByName(name);
    if (i == end()) {
      throw std::domain_error("Invalid id passed to MD_FieldMap::nameToType()");
    }
    return i->type;
  }

}

#endif // __MD_LIB_FIELDMAP_H__
