#ifndef __MD_LIB_MESSAGE_H__
#define __MD_LIB_MESSAGE_H__

#include <algorithm>
#include <iostream>
#include <iterator>
#include <functional>
#include <memory>
#include <string>
#include <vector>
#include <cstring>
#include <set>

#include "MD/private/prefix.h"
#include "MD/MD_sys_types.h"
#include "MD/MD_FieldDefs.h"
#include "MD/MD_Types.h"
#include "MD/MD_DateTime.h"
#include "MD/MD_EpochNS.h"
#include "MD/MD_Variant.h"
#include "MD/MD_FieldMap.h"
#include "MD/MD_Protocols.h"
#include "MD/MD_ReusableContainer.h"
#include "MD/MD_MsgData.h"
#include "MD/MD_Optimizer.h"

namespace MD {

  //! \class MD_Message
  //! \brief The Interface class \b MsgT, IterMap, and other containers
  class MD_LIB_API MD_Message {
  protected:
    // Oladell-approved shorthand.  Much better than 'using namespace.'
    typedef MD::FIELDS::FieldID FieldID;
    typedef MD::DATATYPES::TypeID TypeID;
    typedef MD::MD_Variant MD_Variant;

  public:
    class const_iterator;

    struct Buffer {
      const void * ptr;
      uint16_t size;
    };

    struct String {
      const char * ptr;
      uint16_t size;
    };

    struct Allocator {
        std::shared_ptr<ReusableContainer<MsgData>>   data_container;
        std::shared_ptr<ReusableContainer<Optimizer>> offsets_container;

        Allocator(std::shared_ptr<ReusableContainer<MsgData>>   dc,
                  std::shared_ptr<ReusableContainer<Optimizer>> oc)
            : data_container(dc), offsets_container(oc)
        {}

        Allocator()
            : data_container(new ReusableContainer<MsgData>())
            , offsets_container(new ReusableContainer<Optimizer>())
        {}
    };

    const static Allocator DEFAULT_ALLOCATOR;

    MD_Message() : fmap(MD::MD_FieldMap::instance()), m_data(true), m_pOffsets(true) { }

    MD_Message(const Allocator& allocator)
        : fmap(MD::MD_FieldMap::instance())
        , m_data(true, allocator.data_container)
        , m_pOffsets(true, allocator.offsets_container)
    {}

    virtual ~MD_Message() {}

    virtual MD::PROTOCOLS::MsgProtocol protocol() const = 0;

    virtual const unsigned char *
    getDataPtr (FieldID id, TypeID & type, int & size) const = 0;

    // Give the user a new empty message of the same type as this one.
    // The user is responsible for deleting the pointer.
    virtual std::unique_ptr<MD_Message> create() const = 0;

    // Give the user a new copy of the message.
    // The user is responsible for deleting the pointer.
    virtual std::unique_ptr<MD_Message> clone() const = 0;

    // Give the user a new pointer containing this message.
    // The user is responsible for deleting the pointer.
    virtual std::unique_ptr<MD_Message> move() = 0;

    // createFromPB() returns an auto_ptr to newly allocated MD_Message
    // object of type specified by a protocol byte.  If no MD_Message
    // class corresponds to this protocol byte, createFromPB() returns
    // a NULL pointer.
    static std::unique_ptr<MD_Message> createFromPB(uint8_t protocol_byte);

    // createFromBytes(): Creates a MD_Message object from its serialized
    // byte representation.  Takes buffer (a pointer to a block of
    // memory holding a serialized message) and size (the size of that
    // buffer).  Returns an auto_ptr to a newly allocated MD_Message
    // subclass corresponding to the protocol and contents of the
    // serialized message.  Prefer this function over createFromPB()
    // if possible.
    static std::unique_ptr<MD_Message> createFromBytes(const uint8_t * buffer,
                                                  uint32_t size);
    static std::unique_ptr<MD_Message> createFromBytes(
        const uint8_t* buffer, uint32_t size, const Allocator& allocator);

    // createPreferred() returns an auto_ptr to a newly allocated
    // MD_Message of the currently "preferred" message type, as defined
    // by the message configuration files.  In general,
    // createPreferred() should be used instead of "new MsgT" or "new
    // IterMap".
    static std::unique_ptr<MD_Message> createPreferred(bool check_initial_only = true);

    //! Get a field.
    //! \param id field id to get
    //! \param result parameter in which to place the result
    //! \return true/false to indicate found / not found
    virtual bool getChar     (FieldID id, char &         result) const = 0;
    virtual bool getInt      (FieldID id, int &          result) const = 0;
    virtual bool getInt64    (FieldID id, int64_t &      result) const = 0;
    virtual bool getFloat    (FieldID id, float &        result) const = 0;
    virtual bool getDouble   (FieldID id, double &       result) const = 0;
    virtual bool getString   (FieldID id, String &       result) const = 0;
    virtual bool getBuffer   (FieldID id, Buffer &       result) const = 0;
    virtual bool getDateTime (FieldID id, MD_DateTime &  result) const = 0;
    virtual bool getEpochNS  (FieldID id, MD_EpochNS &   result) const = 0;

    //! Add a new field.  Will not overwrite a field that already exists.
    //! \param id field id to set
    //! \param data value to set
    //! \return true/false to indicate success / failure
    virtual bool addChar     (FieldID id, char         data) = 0;
    virtual bool addInt      (FieldID id, int          data) = 0;
    virtual bool addInt64    (FieldID id, int64_t      data) = 0;
    virtual bool addFloat    (FieldID id, float        data) = 0;
    virtual bool addDouble   (FieldID id, double       data) = 0;
    virtual bool addString   (FieldID id, String       data) = 0;
    virtual bool addBuffer   (FieldID id, Buffer       data) = 0;
    virtual bool addDateTime (FieldID id, MD_DateTime  data) = 0;
    virtual bool addEpochNS  (FieldID id, MD_EpochNS   data) = 0;

    //! Update a field.  Will add if the field does not exist.
    //! \param id field id to set
    //! \param data value to set
    //! \return true/false to indicate success / failure
    virtual bool updateChar     (FieldID id, char        data) = 0;
    virtual bool updateInt      (FieldID id, int         data) = 0;
    virtual bool updateInt64    (FieldID id, int64_t     data) = 0;
    virtual bool updateFloat    (FieldID id, float       data) = 0;
    virtual bool updateDouble   (FieldID id, double      data) = 0;
    virtual bool updateString   (FieldID id, String      data) = 0;
    virtual bool updateBuffer   (FieldID id, Buffer      data) = 0;
    virtual bool updateDateTime (FieldID id, MD_DateTime data) = 0;
    virtual bool updateEpochNS  (FieldID id, MD_EpochNS  data) = 0;

    //! Deletes a field from the message.
    //! \param id Integer value specifying the id of the field to remove
    //! \return true if the field was removed, otherwise false
    virtual bool erase (FieldID id) = 0;

    //! Deletes all the fields in the set from the message
    //! \param set of fids to erase
    //! \return true if all the fields in the set exist and were removed, otherwise false
    virtual bool erase (const std::set<FieldID> &ids) = 0;

    //! Deletes all the fields in the vector from the message
    //! \param vector of fids to erase
    //! \return true if all the fields in the vector exist and were removed, otherwise false
    virtual bool erase (const std::vector<FieldID> &ids) = 0;

    //! resets the ADO for reuse
    virtual void clear() = 0;

    //! Check whether the message has any fields.
    //! \return true if empty, otherwise false
    virtual bool empty() const = 0;

    //! Test for the existance of the specified field in the message.
    //! \param id field to test for
    //! \return true/false for found/not found
    virtual bool exists (FieldID id) const = 0;

    //! Obtain a pointer to the message's binary data buffer.
    //! \return a pointer to the binary data
    virtual const unsigned char* getBinary() const = 0;

    //! Set the binary data of a message.
    //! \param buffer the binary data to set
    //! \param size size of the binary data
    //! \return true
    virtual bool setBinary (const unsigned char* buffer, uint32_t size) = 0;

    //! Obtain size of binary data buffer
    virtual uint32_t getSize() const = 0;

    size_t fieldCount() const { return m_pOffsets->get_field_count(); }

    // non-virtual utility methods
    const_iterator begin() const;
    const_iterator end() const;

    void swap(MD_Message& other);

    bool get (FieldID id, MD_Variant & result) const;
    bool add (FieldID id, const MD_Variant & data);
    bool update (FieldID id, const MD_Variant & data);

    // Get a whole/num/denom value, and automatically convert to a
    // double.  Defaults whole to 0, num to 0, and denom to 100 if
    // these fields cannot be found or are of the wrong type.  Returns
    // false if none of the fields were found, or true if at least one
    // of the fields was found.
    bool getWND (FieldID whole_id, FieldID num_id, FieldID denom_id,
                 double & result) const;

    // Non-virtual convenience accessors for strings.  Because these
    // non-virtual methods have the same names as virtual methods,
    // C++'s hiding rules are an issue.  (See the "C++ FAQs" book for
    // a full explanation, index entry 'hiding rule'.)  To prevent the
    // derived class' accessors from hiding the convenience methods,
    // _every_ MD_Message implementation must explicitly pull in the
    // convenience methods into the derived classes with a 'using'
    // statement, like this:
    // 
    // // Pull in base class convenience methods
    // using MD_Message::getString;
    // using MD_Message::addString;
    // ...
    bool getString (FieldID id, std::string & result) const;
    bool addString (FieldID id, const std::string & data);
    bool addString (FieldID id, const char * data);
    bool addString (FieldID id, const char * data, const uint16_t length);
    bool updateString (FieldID id, const std::string & data);
    bool updateString (FieldID id, const char * data);
    bool updateString (FieldID id, const char * data, const uint16_t length);

    // get/add/update functions without the type in the name, allowing
    // templated functions to interact (without specializing) with MD_Message
    bool get (FieldID id, char &        result) const;
    bool get (FieldID id, int &         result) const;
    bool get (FieldID id, int64_t &     result) const;
    bool get (FieldID id, float &       result) const;
    bool get (FieldID id, double &      result) const;
    bool get (FieldID id, String &      result) const;
    bool get (FieldID id, Buffer &      result) const;
    bool get (FieldID id, MD_DateTime & result) const;
    bool get (FieldID id, MD_EpochNS&   result) const;

    bool add (FieldID id, char        data);
    bool add (FieldID id, int         data);
    bool add (FieldID id, int64_t     data);
    bool add (FieldID id, float       data);
    bool add (FieldID id, double      data);
    bool add (FieldID id, String      data);
    bool add (FieldID id, Buffer      data);
    bool add (FieldID id, MD_DateTime data);
    bool add (FieldID id, MD_EpochNS  data);

    bool update (FieldID id, char        data);
    bool update (FieldID id, int         data);
    bool update (FieldID id, int64_t     data);
    bool update (FieldID id, float       data);
    bool update (FieldID id, double      data);
    bool update (FieldID id, String      data);
    bool update (FieldID id, Buffer      data);
    bool update (FieldID id, MD_DateTime data);
    bool update (FieldID id, MD_EpochNS  data);

    bool get (FieldID id, std::string & result) const;
    bool add (FieldID id, const std::string & data);
    bool add (FieldID id, const char * data);
    bool update (FieldID id, const std::string & data);
    bool update (FieldID id, const char * data);

    // Convert between message types.
    // be advised: this function only adds fields from src message, 
    //             it doesn't remove any fields that aren't in src
    void copyMsg (const MD_Message & src);

    // Print the message to stdout.  Prefer operator<<, please.
    void print() const;

    // don't iterate over an MD_Message with the const_iterator object 
    // unless you intend to get every field value in the message and do 
    // so via reference to *iterator. Use this function instead.
    void get_field_ids(std::vector<MD::FIELDS::FieldID> & fields) const;

  protected:
    template <class FieldType, class InputIterator>
    bool bulkErase(InputIterator first, InputIterator last, const size_t num_fields);

    // Optimize the message starting at a point in the message.
    template <class FieldType>
    void optimize (MsgData::iterator i);
    
    template <class FieldType>
    FieldType getField(FieldID id) const;

    void printError(FieldID id, TypeID type, size_t size) const;

    const MD_FieldMap                         & fmap;
    ReusableContainerWrapper<MsgData>           m_data;
    mutable ReusableContainerWrapper<Optimizer> m_pOffsets;

  private:
    virtual void getFieldIDs(std::back_insert_iterator<std::vector<MD::FIELDS::FieldID> > inserter) const = 0;

    // Do not allow implicit conversions between message types.
    MD_Message & operator= (const MD::MD_Message & source);
  };

  // This class is for the convenience of printing an MD::MD_Message with
  // the WHOLE NUM DENOM fields converted to a float.
  // Usage: cout << "Here is my message: " << endl << PriceOutput(msg);
  class PriceOutput {
  public:
    PriceOutput(const MD::MD_Message & msg) : m_msg(msg) {}
    const MD::MD_Message & m_msg;
  };

  template <class FieldType>
  struct FieldOffsetPair
  {
     FieldOffsetPair(FieldType fld, size_t off) : field(fld), offset(off) { }

     bool operator>(const FieldOffsetPair& rhs) const
     {
        return offset > rhs.offset;
     }

     FieldType field;
     size_t offset;
  };
}


// MD_Message's iterator declarations

class MD::MD_Message::const_iterator {
private:
  // shorthand only
  typedef MD::FIELDS::FieldID FieldID;
  typedef MD::MD_Variant MD_Variant;

public:
  typedef std::pair <FieldID, MD_Variant> value_type;

  const_iterator();
  // default copy constructor should be OK
  const_iterator & operator= (const_iterator src);
  const_iterator & operator++();
  const_iterator operator++ (int);
  bool operator!= (const const_iterator & src) const;
  bool operator== (const const_iterator & src) const;
  const value_type & operator*() const;
  const value_type * operator->() const;
  FieldID getId() const;

private:
  friend class MD::MD_Message;

  const_iterator(const MD::MD_Message & msg);
  void updateval();

  const MD::MD_Message * m_msg;
  std::vector<FieldID> m_ids;   // field ids to iterate across
  uint32_t m_pos;               // current position in vector of field ids
  value_type m_result;
};


namespace MD {
// Stand-alone function declarations
MD_LIB_API std::ostream & operator << (std::ostream & os, const MD::MD_Message & msg);
MD_LIB_API std::ostream & operator << (std::ostream & os, const MD::PriceOutput & price_output);
}

// Inlined non-virtual member definitions

inline void MD::MD_Message::const_iterator::updateval() {
  m_result.first = m_ids[m_pos];
  if (m_result.first != 65535) {
#ifndef NDEBUG
    bool b =
#endif
      m_msg->get(m_result.first, m_result.second);
    assert(b == true);
  }
}

inline MD::MD_Message::const_iterator MD::MD_Message::begin() const {
  const_iterator i = const_iterator(*this);
  std::back_insert_iterator<std::vector<MD::FIELDS::FieldID> > inserter(i.m_ids);
  getFieldIDs(inserter);
  i.m_ids.push_back(65535);
  i.m_pos = 0;
  i.updateval();

  return i;
}

inline MD::MD_Message::const_iterator MD::MD_Message::end() const {
  const_iterator i = const_iterator(*this);
  i.m_ids.push_back(65535);
  i.m_pos = 0;
  return i;
}

inline void MD::MD_Message::swap(MD::MD_Message& other) {
    m_data.swap(other.m_data);
    m_pOffsets.swap(other.m_pOffsets);
}

inline void MD::MD_Message::get_field_ids(std::vector<MD::FIELDS::FieldID> & fields) const {
  fields.clear();
  std::back_insert_iterator<std::vector<MD::FIELDS::FieldID> > inserter(fields);
  getFieldIDs(inserter);
}

inline bool MD::MD_Message::getString (FieldID id, std::string & result) const {
  String s;
  bool b = getString(id, s);
  if (b) { result.assign(s.ptr, s.size); }
  return b;
}

inline bool MD::MD_Message::addString (FieldID id, const char * data, const uint16_t length) {
  const String s = { data, length };
  return addString(id, s);
}
inline bool MD::MD_Message::addString (FieldID id, const std::string & data) {
  return addString(id, data.data(), static_cast<uint16_t>(data.size()));
}
inline bool MD::MD_Message::addString (FieldID id, const char * data) {
  return addString(id, data, static_cast<uint16_t>(strlen(data)));
}

inline bool MD::MD_Message::updateString (FieldID id, const char * data, const uint16_t length) {
  const String s = { data, length };
  return updateString(id, s);
}
inline bool MD::MD_Message::updateString (FieldID id, const std::string & data) {
  return updateString(id, data.data(), static_cast<uint16_t>(data.size())); 
}
inline bool MD::MD_Message::updateString (FieldID id, const char * data) {
  return updateString(id, data, static_cast<uint16_t>(strlen(data)));
}

inline bool MD::MD_Message::get (FieldID id, char & result) const { return getChar(id, result); }
inline bool MD::MD_Message::get (FieldID id, int & result) const { return getInt(id, result); }
inline bool MD::MD_Message::get (FieldID id, int64_t & result) const { return getInt64(id, result); }
inline bool MD::MD_Message::get (FieldID id, float & result) const { return getFloat(id, result); }
inline bool MD::MD_Message::get (FieldID id, double & result) const { return getDouble(id, result); }
inline bool MD::MD_Message::get (FieldID id, String & result) const { return getString(id, result); }
inline bool MD::MD_Message::get (FieldID id, Buffer & result) const { return getBuffer(id, result); }
inline bool MD::MD_Message::get (FieldID id, MD_DateTime & result) const { return getDateTime(id, result); }
inline bool MD::MD_Message::get (FieldID id, MD_EpochNS & result) const { return getEpochNS(id, result); }

inline bool MD::MD_Message::add (FieldID id, char data) { return addChar(id, data); } 
inline bool MD::MD_Message::add (FieldID id, int data) { return addInt(id, data); }
inline bool MD::MD_Message::add (FieldID id, int64_t data) { return addInt64(id, data); }
inline bool MD::MD_Message::add (FieldID id, float data) { return addFloat(id, data); }
inline bool MD::MD_Message::add (FieldID id, double data) { return addDouble(id, data); }
inline bool MD::MD_Message::add (FieldID id, String data) { return addString(id, data); }
inline bool MD::MD_Message::add (FieldID id, Buffer data) { return addBuffer(id, data); }
inline bool MD::MD_Message::add (FieldID id, MD_DateTime data) { return addDateTime(id, data); }
inline bool MD::MD_Message::add (FieldID id, MD_EpochNS data) { return addEpochNS(id, data); }

inline bool MD::MD_Message::update (FieldID id, char data) { return updateChar(id, data); }
inline bool MD::MD_Message::update (FieldID id, int data) { return updateInt(id, data); }
inline bool MD::MD_Message::update (FieldID id, int64_t data) { return updateInt64(id, data); }
inline bool MD::MD_Message::update (FieldID id, float data) { return updateFloat(id, data); }
inline bool MD::MD_Message::update (FieldID id, double data) { return updateDouble(id, data); }
inline bool MD::MD_Message::update (FieldID id, String data) { return updateString(id, data); }
inline bool MD::MD_Message::update (FieldID id, Buffer data) { return updateBuffer(id, data); }
inline bool MD::MD_Message::update (FieldID id, MD_DateTime data) { return updateDateTime(id, data); }
inline bool MD::MD_Message::update (FieldID id, MD_EpochNS data) { return updateEpochNS(id, data); }

inline bool MD::MD_Message::get (FieldID id, std::string & result) const { return getString(id, result); }
inline bool MD::MD_Message::add (FieldID id, const std::string & data) { return addString(id, data); }
inline bool MD::MD_Message::add (FieldID id, const char * data) { return addString(id, data); }
inline bool MD::MD_Message::update (FieldID id, const std::string & data) { return updateString(id, data); }
inline bool MD::MD_Message::update (FieldID id, const char * data) { return updateString(id, data); }


// NOTE: This function could run faster if implemented in derived
// classes, instead of the parent class.  If the performance of
// print() is an issue, however, you've probably got other issues to
// worry about.
inline void MD::MD_Message::print() const {
  // std::cout << *this << std::endl;
}

template<class FieldType>
inline FieldType MD::MD_Message::getField(FieldID id) const {
  size_t offset = m_pOffsets->get(id);
  return (offset != 0) ? FieldType(m_data->begin() + offset) : FieldType(NULL);
}

template<class FieldType>
void MD::MD_Message::optimize(MsgData::iterator i)
{
  bool print_msg = false;
  static int printnum = 0;

  while (i < m_data->end()) {
    FieldType field(i);

    // If the incoming message has bad fields, we should log the
    // contents of the bad field, throw out the field, and print the
    // (good) remainder of the message, for reference.  We use a
    // function scoped static to limit logging to 100 times, in case
    // we are deluged by bad field ids.
    // TODO: Nuke the logging limit, once bad field ids become a rare
    // occurrence.  Alternately, use logger.
    while (!m_pOffsets->set(field.id(), field.type(), i - m_data->begin())) {
      if (printnum <= 100) {
        std::cerr << "Error: Bad Field on wire!  Skipping!" << std::endl;
        printError(field.id(), field.type(), field.size());
        printnum++;
      }

      m_data->erase(i, (std::min)(field.next_ptr(), m_data->end()));
      print_msg = true;
      if (i >= m_data->end()) { break; }
    }

    i = field.next_ptr();
  }

  if (print_msg && printnum <= 100) {
    std::cerr << "Remainder of message after stripping fields:" << std::endl
              << "  " << *this << std::endl;
  }
}

template<class FieldType, class InputIterator>
bool MD::MD_Message::bulkErase(InputIterator first, InputIterator last, const size_t num_fields)
{
  if (first == last || num_fields==0)
    return true;

  if (empty())
    return false;

  typedef std::vector<FieldOffsetPair<FieldType> > FieldOffsetVector_t;
  FieldOffsetVector_t field_offset;
  field_offset.reserve(num_fields);

  while (first != last) {
    FieldType field(getField<FieldType>(*first));

    if (field.isValid())
      field_offset.push_back(FieldOffsetPair<FieldType>(field, m_pOffsets->get(*first)));

    ++first;
  }

  if (field_offset.size() == 0)
    return false;

  std::sort(field_offset.begin(), field_offset.end(), std::greater<FieldOffsetPair<FieldType> >()); 

  typename FieldOffsetVector_t::const_iterator cur(field_offset.begin()), end(field_offset.end()), temp;

  for (; cur != end; ++cur) {
    temp = cur;

    m_pOffsets->erase(cur->field.id());

    // skip duplicate fids in the input (if vector flavor is used). Use the last instance from all the dups
    while ((temp != end - 1) && ((temp + 1)->field.id() == temp->field.id()))
      ++temp;

    // contiguous field deletes with one memmove
    while ((temp != end - 1) && ((temp + 1)->field.next_ptr() == temp->field.ptr()))
    {
      ++temp;
      // there can be duplicates of fid1 followed duplicates of fid2, use the last copy of fid2
      while ((temp != end - 1) && ((temp + 1)->field.id() == temp->field.id()))
        ++temp;
      
      m_pOffsets->erase(temp->field.id());
    }
    
    m_data->erase(temp->field.ptr(), cur->field.next_ptr());
    cur = temp;
  }

  optimize<FieldType>(field_offset.back().field.ptr());

  return (field_offset.size() == num_fields);
}

inline MD::MD_Message::const_iterator::const_iterator()
  : m_msg(NULL)
  , m_ids()
  , m_pos(0)
  , m_result() { }

inline MD::MD_Message::const_iterator::const_iterator(const MD::MD_Message & msg)
  : m_msg(&msg)
  , m_ids()
  , m_pos(0)
  , m_result() { }


inline MD::MD_Message::const_iterator & MD::MD_Message::const_iterator::operator= (const_iterator src) {
  std::swap(m_msg, src.m_msg);
  std::swap(m_ids, src.m_ids);
  std::swap(m_pos, src.m_pos);
  std::swap(m_result, src.m_result);
  return *this;
}

inline MD::MD_Message::const_iterator & MD::MD_Message::const_iterator::operator++() {
  ++m_pos;
  updateval();
  return *this;
}

inline MD::MD_Message::const_iterator MD::MD_Message::const_iterator::operator++ (int) {
  const_iterator copy(*this);
  ++m_pos;
  updateval();
  return copy;
}

inline bool MD::MD_Message::const_iterator::operator!= (const const_iterator & src) const {
  return m_ids[m_pos] != src.m_ids[src.m_pos];
}

inline bool MD::MD_Message::const_iterator::operator== (const const_iterator & src) const {
  return m_ids[m_pos] == src.m_ids[src.m_pos];
}

inline const MD::MD_Message::const_iterator::value_type & MD::MD_Message::const_iterator::operator*() const {
  return m_result;
}

inline const MD::MD_Message::const_iterator::value_type * MD::MD_Message::const_iterator::operator->() const {
  return & operator*();
}

inline MD::FIELDS::FieldID MD::MD_Message::const_iterator::getId() const {
  return m_ids[m_pos];
}

namespace MD {
template<class MD_MessageType>
class MD_MsgIterator
{
  public:
  typedef typename MD_MessageType::Field value_type;

  MD_MsgIterator(const unsigned char *ptr) : m_field((char *)ptr) {}
  MD_MsgIterator(const char          *ptr = NULL) : m_field((char *)ptr) {}

  // These comparison operations are too dangerous when trying to detect
  // the end of a buffer.  Use operator< instead.
  // bool operator==(const MD_MsgIterator &rhs) const
  // { return rhs.m_field.ptr() == m_field.ptr(); }
  // bool operator!=(const MD_MsgIterator &rhs) const
  // { return rhs.m_field.ptr() != m_field.ptr(); }

  bool operator<(const MD_MsgIterator &rhs) const
  { return m_field.ptr() < rhs.m_field.ptr(); }

  MD_MsgIterator operator++()
  { m_field.next(); return *this; }

  MD_MsgIterator operator++(int)
  { MD_MsgIterator tmp(*this); m_field.next(); return tmp; }

  const value_type &operator*()  const { return  m_field; }
  const value_type *operator->() const { return &m_field; }

  private:
    value_type m_field;
};
}


#endif // __MD_LIB_MESSAGE_H__
