#ifndef __MD_LIB_MSG_DATA_H__
#define __MD_LIB_MSG_DATA_H__

#include <algorithm>
#include <iostream>

#include <cassert>
#include <cstdlib>
#include <string.h>

namespace MD {
class MsgData {
public:
  typedef size_t size_type;
  typedef char value_type;
  typedef value_type * iterator;
  typedef const value_type * const_iterator;

  MsgData();

  template <typename InputIterator>
  MsgData (InputIterator first, InputIterator last);

  MsgData (const MsgData & other);

  ~MsgData();

  MsgData & operator= (const MsgData & other);

  // swaps this buffer with that of another MsgData
  void swap (MsgData & other);

  iterator begin();
  const_iterator begin() const;
  iterator end();
  const_iterator end() const;

  void clear();
  bool empty() const;

  // Inserts the range [first, last) before pos, just like std::vector.
  template <typename InputIterator>
  void insert (iterator pos, InputIterator first, InputIterator last);

  // Erases the range [first, last)
  iterator erase(iterator first, iterator last);

  size_type size() const;
  size_type capacity() const;
  void reserve (size_type size);
  void resize (size_type size);

private:
  // BLOCK_SIZE must be a power of 2
  // BLOCK_MASK must be BLOCK_SIZE - 1
  enum { BLOCK_MASK = 0x3ff, BLOCK_SIZE = 0x400 };

  value_type * m_buffer;
  size_type m_capacity;
  size_type m_size;
};


inline MsgData::MsgData()
  : m_buffer(NULL)
  , m_capacity(0)
    , m_size(0) { }

inline MsgData::iterator MsgData::begin() {
  return m_buffer;
}

inline MsgData::const_iterator MsgData::begin() const {
  return m_buffer;
}

inline void MsgData::reserve (MsgData::size_type size) {
  if (size > m_capacity) {
    MsgData::size_type new_capacity((size + BLOCK_MASK) & ~BLOCK_MASK);
    m_buffer = reinterpret_cast <MsgData::value_type *>
      (realloc(m_buffer, new_capacity));
    m_capacity = new_capacity;
  }
}

inline MsgData::iterator MsgData::end() {
  return m_buffer + m_size;
}

inline MsgData::const_iterator MsgData::end() const {
  return m_buffer + m_size;
}

template <typename InputIterator>
inline void MsgData::insert (MsgData::iterator pos,
                             InputIterator first,
                             InputIterator last) {
  assert(last >= first);
  assert(pos >= begin());
  assert(pos <= end());

  size_type pos_offset = pos - begin(); // pos as an offset from begin()
  size_type size_of_range = last - first;
  size_type new_size = m_size + size_of_range;

  reserve(new_size);
  pos = begin() + pos_offset;  // since reserve() invalidates pos

  // If we are inserting in to the middle of the buffer, we need to
  // make room for the insertion by memmove()ing the second half of
  // the buffer.
  if (pos != end()) {
    memmove(pos + size_of_range, pos, m_size - pos_offset);
  }

  // Now that room has been made, copy the range into the buffer.
  memcpy(pos, first, size_of_range);

  m_size = new_size;
}

template <typename InputIterator>
inline MsgData::MsgData (InputIterator first, InputIterator last)
  : m_buffer(NULL)
  , m_capacity(0)
  , m_size(0) {
  insert(begin(), first, last);
}

inline void MsgData::clear() {
  m_size = 0;
}

inline MsgData & MsgData::operator= (const MsgData & other) {
  if (this == &other) return *this; // handle self-assignment
  clear();
  insert(begin(), other.begin(), other.end());
  return *this;
}

inline MsgData::MsgData (const MsgData & other)
  : m_buffer(NULL)
  , m_capacity(0)
  , m_size(0) {
  *this = other;
}

inline MsgData::~MsgData() {
  free(m_buffer);
}

inline void MsgData::swap (MsgData & other) {
  std::swap(m_buffer, other.m_buffer);
  std::swap(m_capacity, other.m_capacity);
  std::swap(m_size, other.m_size);
}

inline bool MsgData::empty() const {
  return (0 == m_size);
}

inline MsgData::iterator MsgData::erase (MsgData::iterator first,
                                         MsgData::iterator last) {
  assert(last >= first);
  assert(first >= begin());
  assert(last <= end());

  memmove(first, last, end() - last);
  m_size -= last - first;
  return first;
}


inline MsgData::size_type MsgData::size() const {
  return m_size;
}

inline MsgData::size_type MsgData::capacity() const {
  return m_capacity;
}

inline void MsgData::resize (MsgData::size_type size) {
  reserve(size);
  m_size = size;
}
}

#endif // __MD_LIB_MSG_DATA_H__
